// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/identifier.dart';

abstract class Event extends Identifiable<EventId> implements JsonWritable {
  String get _type;

  final EventId id;

  Event(this.id);

  factory Event.fromJson(RoomId roomId, Map<String, dynamic> json) {
    var type = json['type'];

    // Skip if it's not a valid event id
    if (!EventId.isValidFullyQualified(json['event_id'])) {
      return null;
    }

    // These might be null
    var id = EventId(json['event_id']);
    var sender = UserId(json['sender']);

    var originServerTs = json['origin_server_ts'];
    var time;
    if (originServerTs != null) {
      time = DateTime.fromMillisecondsSinceEpoch(originServerTs);
    }

    switch (type) {
      case NameChangeEvent.type:
        return NameChangeEvent.fromJson(id, roomId, sender, time, json);
      case AvatarChangeEvent.type:
        return AvatarChangeEvent.fromJson(id, roomId, sender, time, json);
      case MessageEvent.type:
        return MessageEvent.fromJson(id, roomId, sender, time, json);
      default:
        return null;
    }
  }

  @override
  Map<String, dynamic> toJson() => {
    'event_id': id.toJson(),
    'type': _type
  };

}

abstract class RoomEvent extends Event {
  final RoomId roomId;

  final UserId sender;
  final DateTime time;

  RoomEvent(EventId id, this.roomId, this.sender, this.time) : super(id);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'sender': sender.toJson(),
      'origin_server_ts': time.millisecondsSinceEpoch
    });
}

abstract class StateEvent extends RoomEvent {
  final String stateKey = '';

  StateEvent(EventId id, RoomId roomId, UserId sender, DateTime time)
      : super(id, roomId, sender, time);

  @override
  Map<String, dynamic> toJson() =>
      super.toJson()..addAll({'state_key': stateKey});
}

class NameChangeEvent extends StateEvent {
  static const String type = 'm.room.name';

  @override
  final String _type = type;

  final String name;

  NameChangeEvent(
      this.name, EventId id, RoomId roomId, UserId sender, DateTime time)
      : super(id, roomId, sender, time);

  factory NameChangeEvent.fromJson(
      EventId id,RoomId roomId, UserId sender,
      DateTime time, dynamic json) {
    var name = json['content']['name'];
    return NameChangeEvent(name, id, roomId, sender, time);
  }

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'type': type,
      'content': {'name': name}
    });
}

class AvatarChangeEvent extends StateEvent {
  static const String type = 'm.room.avatar';

  @override
  final String _type = type;

  /// An `mxc` url pointing to the avatar.
  final Uri url;

  AvatarChangeEvent(
      this.url, EventId id, RoomId roomId, UserId sender, DateTime time)
      : super(id, roomId, sender, time);

  factory AvatarChangeEvent.fromJson(
      EventId id,RoomId roomId, UserId sender,
      DateTime time, dynamic json) {
    var url = json['content']['url'];
    return AvatarChangeEvent(Uri.parse(url), id, roomId, sender, time);
  }

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'type': type,
      'content': {'url': url}
    });
}

abstract class MessageEvent extends RoomEvent {
  static const String type = 'm.room.message';

  @override
  final String _type = type;

  String get _messageType;

  MessageEvent(EventId id, RoomId roomId, UserId sender, DateTime time)
      : super(id, roomId, sender, time);

  factory MessageEvent.fromJson(EventId id, RoomId roomId, UserId sender,
      DateTime time, Map<String, dynamic> json) {
    var msgtype = json['content']['msgtype'];

    switch (msgtype) {
      case TextMessageEvent.type:
      default:
        var body = json['content']['body'];
        return TextMessageEvent(body, id, roomId, sender, time);
    }
  }

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'content': {
        'msgtype': _messageType
      }
    }
  );
}

class TextMessageEvent extends MessageEvent {
  static const String type = 'm.text';

  @override
  final String _messageType = type;

  final String body;

  TextMessageEvent(this.body, EventId id, RoomId roomId, UserId sender, DateTime time)
      : super(id, roomId, sender, time);

  @override
  Map<String, dynamic> toJson() => super.toJson()
  ..['content']
    .addAll({
      'body': body
    }
  );
}